@@ -9,10 +9,7 @@
 # directory, you need to use --prefix=<dir>, or run with root priviledges.
 
 if [ -d ../wisi-3.1.? ]; then
-    WISI_DIR=`ls -d ../wisi-3.1.?`
-else
-    # try devel version
-    WISI_DIR=`ls -d ../wisi-3.1.?.0.*`
+    WISI_DIR=`ls -d ../wisi-4.0.?`
 fi
 
 gprinstall -f -p -P ada_mode_wisi_parse.gpr -aP $WISI_DIR --install-name=ada_mode_wisi_parse $1 
